from __future__ import annotations
from typing import TYPE_CHECKING, TypeAlias
import itertools
import logging

import claripy

import angr
from angr.engines.ail.callstack import AILCallStack
from angr.engines.light.engine import SimEngineLightAIL
from angr import ailment, errors
from angr.engines.successors import SimSuccessors
from angr.sim_state import SimState
from angr.engines.vex.claripy import ccall
from angr.storage.memory_mixins.memory_mixin import MemoryMixin
from angr.utils.constants import DEFAULT_STATEMENT

if TYPE_CHECKING:
    from angr.project import Project
    from angr.analyses.decompiler.clinic import Clinic

log = logging.getLogger(__name__)

StateType: TypeAlias = SimState[ailment.Address, ailment.Address]
DataType: TypeAlias = claripy.ast.Bits | claripy.ast.Bool


class CallReached(Exception):
    """
    An exception to abort executing a block if we need to restart it with a call result
    """


class SimEngineAILSimState(SimEngineLightAIL[StateType, DataType, bool, None]):
    """
    A light engine for symbolically executing AIL
    """

    def __init__(
        self,
        project: Project,
        successors: SimSuccessors,
    ):
        self.successors = successors
        self.ret_idx = 0
        super().__init__(project)

    def process(
        self, state: StateType, *, block: ailment.Block | None = None, whitelist: set[int] | None = None, **kwargs
    ) -> None:
        self.state = state
        self.state.bbl_addr = state.addr[0]
        # if there is a function parameter handoff waiting, process that asap
        if self.frame.passed_args is not None:
            clinic = self.lift_addr(state.addr)
            assert clinic.arg_vvars is not None
            expected = len(clinic.arg_vvars)
            got = len(self.frame.passed_args)
            if got < expected:
                raise errors.AngrRuntimeError(
                    f"Function entry missing args: expected={expected} got={got} at {state.addr}"
                )
            if got > expected:
                log.debug("Function entry extra args: expected=%d got=%d at %s", expected, got, state.addr)
            for idx, value in enumerate(self.frame.passed_args):
                if idx >= len(clinic.arg_vvars):
                    break
                vvar, _ = clinic.arg_vvars[idx]
                self._do_assign(vvar, value, auto_narrow=True)
            self.frame.passed_args = None

        if self.frame.resume_at is not None:
            if block is None:
                block = self.lift(state)
            whitelist2 = set(range(self.frame.resume_at, len(block.statements)))
            if whitelist is None:
                whitelist = whitelist2
            else:
                whitelist.intersection_update(whitelist2)
        super().process(state, block=block, whitelist=whitelist, **kwargs)

    @property
    def frame(self) -> AILCallStack:
        callstack = self.state.callstack
        assert isinstance(callstack, AILCallStack)
        return callstack

    def lift_addr(self, addr: ailment.Address) -> Clinic:
        result = self.state.globals["ail_lifter"]  # type: ignore
        assert callable(result)
        return result(addr[0])  # type: ignore

    def lift(self, state: StateType | int | ailment.Address) -> ailment.Block:
        addr = (state, None) if isinstance(state, int) else state if isinstance(state, tuple) else state.addr
        clinic = self.lift_addr(addr)
        assert clinic.cc_graph is not None
        blocks = [blk for blk in clinic.cc_graph if (blk.addr, blk.idx) == addr]
        if len(blocks) == 0:
            raise errors.AngrLifterError("Lifted graph does not have the needed block")
        if len(blocks) > 1:
            raise errors.AngrLifterError("Lifted graph contains more than one of the needed block")
        return blocks[0]

    def _top(self, bits):
        return claripy.BVS("ail_engine_top", bits)

    def _is_top(self, expr):
        return expr.op == "BVS" and expr.args[0].name.startswith("ail_engine_top")

    def _find_ptr_region(self, ptr: claripy.ast.BV) -> tuple[MemoryMixin, claripy.ast.BV | int]:
        region: MemoryMixin | None = None
        offset = 0
        queue = [ptr]
        while queue:
            node = queue.pop()
            if node.op == "__add__":
                queue.extend(node.args)  # type: ignore
            elif node.op == "__sub__":
                queue.append(node.args[0])  # type: ignore
                queue.extend(-x for x in node.args[1:])  # type: ignore
            elif node.op == "BVS":
                frame = self.frame
                while frame is not None:
                    referred = frame.var_refs.get(node, None)
                    if referred is None:
                        frame = frame.next
                        continue
                    if region is None:
                        _region = frame.vars[referred]
                        assert isinstance(_region, MemoryMixin)
                        region = _region
                    else:
                        log.warning("Emulation is adding together two pointers")
                        return self.state.memory, ptr
                    break
                else:
                    offset += node
            else:
                offset += node

        if region is None:
            return self.state.memory, ptr
        return region, offset

    def _stmt_diverges(self, result):
        return not result

    def _process_block_end(self, block: ailment.Block, stmt_data: list[bool], whitelist: set[int] | None) -> None:
        if all(stmt_data):
            # this is a block lifted such that we don't have explicit gotos. Find one from the graph
            clinic = self.lift_addr(self.state.addr)
            assert clinic.cc_graph is not None
            if len(clinic.cc_graph.succ[block]) > 1:
                raise errors.AngrRuntimeError(
                    f"Reached default exit of block with {len(clinic.cc_graph.succ[block])} successors"
                )
            if len(clinic.cc_graph.succ[block]) == 0:
                # deadend. add pathterminator
                self.successors.add_successor(
                    self.state,
                    self.state.addr,
                    claripy.true(),
                    "Ijk_Exit",
                    add_guard=False,
                    exit_ins_addr=self.ins_addr,
                    exit_stmt_idx=DEFAULT_STATEMENT,
                )
                return
            succ = next(iter(clinic.cc_graph.succ[block]))
            self.successors.add_successor(
                self.state,
                (succ.addr, succ.idx),
                claripy.true(),
                "Ijk_Boring",
                add_guard=False,
                exit_ins_addr=self.ins_addr,
                exit_stmt_idx=DEFAULT_STATEMENT,
            )

    def _stmt(self, stmt: ailment.statement.Statement, toplevel: bool = True) -> bool:
        if toplevel:
            self.ret_idx = 0
            self.state.scratch.stmt_idx = self.stmt_idx
            self.state.scratch.ins_addr = self.ins_addr
        try:
            result = super()._stmt(stmt)
        except CallReached as e:
            if not toplevel:
                raise errors.AngrRuntimeError(
                    "there is absolutely no good way to emulate this. generate better IR."
                ) from e
            # assume we've already pushed the callee frame
            # furthermore assume that it's okay to mutate callstack.next since it should be unique still...?
            assert self.frame.next is not None
            self.frame.next.resume_at = self.stmt_idx
            return False
        else:
            if toplevel:
                self.frame.passed_rets = ()
                self.frame.resume_at = None
            return result

    def _expr_bool(self, expr) -> claripy.ast.Bool:
        result = self._expr(expr)
        if isinstance(result, claripy.ast.BV) and len(result) == 1:
            result = result != 0
        assert isinstance(result, claripy.ast.Bool)
        return result

    def _expr_bits(self, expr) -> claripy.ast.Bits:
        result = self._expr(expr)
        assert isinstance(result, claripy.ast.Bits)
        return result

    def _expr_bv(self, expr) -> claripy.ast.BV:
        result = self._expr(expr)
        assert isinstance(result, claripy.ast.BV)
        return result

    def _expr_fp(self, expr) -> claripy.ast.FP:
        result = self._expr(expr)
        assert isinstance(result, claripy.ast.FP)
        return result

    def _do_call(self, call: ailment.statement.Call, is_expr: bool = False):
        if angr.options.CALLLESS in self.state.options:
            if is_expr:
                # ????? if doing ret emulation and this is an expr (no lvalue expression)
                # how do I tell if this is a float ret or not?
                return (claripy.BVS(f"callless_stub_{call.target}", call.bits),)
            if call.ret_expr is not None:
                return (claripy.BVS(f"callless_stub_{call.target}", call.ret_expr.bits),)
            if call.fp_ret_expr is not None:
                return (
                    claripy.FPS(
                        f"callless_stub_{call.target}",
                        claripy.FSORT_FLOAT if call.fp_ret_expr.bits == 32 else claripy.FSORT_DOUBLE,
                    ),
                )
            return ()
        arguments = tuple(self._expr_bits(e) for e in (call.args or []))
        target_addr = self._expr_bv(call.target)
        assert target_addr.concrete
        if self.ret_idx < len(self.frame.passed_rets):
            results = self.frame.passed_rets[self.ret_idx]
            self.ret_idx += 1
            return results
        target = (target_addr.concrete_value, None)

        new_frame = AILCallStack(func_addr=target)
        new_frame.passed_args = arguments
        new_frame.return_addr = self.state.addr
        self.frame.push(new_frame)

        self.successors.add_successor(
            self.state,
            target,
            claripy.true(),
            "Ijk_Call",
            add_guard=False,
            exit_ins_addr=self.ins_addr,
            exit_stmt_idx=self.stmt_idx,
        )
        raise CallReached

    def _do_assign(self, dst: ailment.Expression, val: claripy.ast.Bits, auto_narrow: bool = False):
        if len(val) != dst.bits:
            if auto_narrow and len(val) > dst.bits and isinstance(val, claripy.ast.BV):
                val = val[dst.bits - 1 : 0]
            else:
                raise errors.AngrRuntimeError(f"Bad-sized assignment: expected {dst.bits} bits, got {len(val)}")
        match dst:
            case ailment.expression.VirtualVariable():
                if isinstance((mem := self.frame.vars.get(dst.varid, None)), MemoryMixin):
                    mem.store(0, val, endness=self.state.arch.memory_endness)
                else:
                    self.frame.vars[dst.varid] = val
            case ailment.expression.Register():
                self.state.registers.store(dst.reg_offset, val)
            case ailment.expression.Tmp():
                self.tmps[dst.tmp_idx] = val
            case _:
                assert False, f"Unsupported type of assignemnt dst {type(dst)}"

    def _handle_stmt_Assignment(self, stmt) -> bool:
        val = self._expr_bits(stmt.src)
        self._do_assign(stmt.dst, val)
        return True

    def _handle_stmt_Store(self, stmt: ailment.statement.Store) -> bool:
        val = self._expr(stmt.data)
        ptr = self._expr_bv(stmt.addr)
        region, offset = self._find_ptr_region(ptr)
        region.store(offset, val, endness=stmt.endness)
        return True

    def _handle_stmt_WeakAssignment(self, stmt: ailment.statement.WeakAssignment) -> bool:
        raise NotImplementedError(ailment.statement.WeakAssignment)

    def _handle_stmt_CAS(self, stmt: ailment.statement.CAS) -> bool:
        raise NotImplementedError(ailment.statement.CAS)

    def _handle_stmt_Jump(self, stmt: ailment.statement.Jump) -> bool:
        target_addr = self._expr_bv(stmt.target)
        assert target_addr.concrete
        target = (target_addr.concrete_value, stmt.target_idx)
        self.successors.add_successor(
            self.state,
            target,
            claripy.true(),
            "Ijk_Boring",
            add_guard=False,
            exit_ins_addr=self.ins_addr,
            exit_stmt_idx=self.stmt_idx,
        )
        return False

    def _handle_stmt_ConditionalJump(self, stmt: ailment.statement.ConditionalJump) -> bool:
        condition = self._expr_bool(stmt.condition)
        state_true = self.state.copy()
        state_false = self.state
        target_true_addr = self._expr_bv(stmt.true_target)
        assert target_true_addr.concrete
        if stmt.false_target is None:
            target_false_addr = None
        else:
            target_false_addr = self._expr_bv(stmt.false_target)
            assert target_false_addr.concrete
        self.successors.add_successor(
            state_true,
            (target_true_addr.concrete_value, stmt.true_target_idx),
            condition,
            "Ijk_Boring",
            exit_stmt_idx=self.stmt_idx,
            exit_ins_addr=self.ins_addr,
        )
        if target_false_addr is not None:
            self.successors.add_successor(
                state_false,
                (target_false_addr.concrete_value, stmt.false_target_idx),
                ~condition,  # type: ignore
                "Ijk_Boring",
                exit_stmt_idx=self.stmt_idx,
                exit_ins_addr=self.ins_addr,
            )
            return False
        return True

    def _handle_stmt_Return(self, stmt: ailment.statement.Return) -> bool:
        ret_values = tuple(self._expr_bits(e) for e in stmt.ret_exprs)
        target = self.frame.return_addr
        this_frame = self.frame
        # store vvars if needed
        # TODO move this to the history plugin
        if this_frame.vars:
            if "vvars" not in self.state.globals:
                self.state.globals["vvars"] = {}
            if this_frame.func_addr not in self.state.globals["vvars"]:
                self.state.globals["vvars"][this_frame.func_addr] = []
            self.state.globals["vvars"][this_frame.func_addr].append(this_frame.vars)
        self.frame.pop()
        self.frame.passed_rets += (ret_values,)
        self.successors.add_successor(
            self.state,
            target if target is not None else self.state.addr,
            claripy.true(),
            "Ijk_Ret" if target is not None else "Ijk_Exit",
            add_guard=False,
            exit_ins_addr=self.ins_addr,
            exit_stmt_idx=self.stmt_idx,
        )
        return False

    def _handle_stmt_DirtyStatement(self, stmt: ailment.statement.DirtyStatement) -> bool:
        raise NotImplementedError(ailment.statement.DirtyStatement)

    def _handle_stmt_Label(self, stmt: ailment.statement.Label) -> bool:
        return True

    def _handle_stmt_Call(self, stmt: ailment.statement.Call) -> bool:
        results = self._do_call(stmt)
        ret_expr = stmt.ret_expr or stmt.fp_ret_expr
        ret_exprs = [] if ret_expr is None else [ret_expr]
        if len(results) < len(ret_exprs):
            raise errors.AngrRuntimeError(
                f"Call statement expects {len(ret_exprs)} return value(s) but called function provided {len(results)}"
            )
        for ret_expr, result in zip(ret_exprs, results):
            # these may be provided by misbehaving simprocedures. truncate the result as needed.
            self._do_assign(ret_expr, result, auto_narrow=True)
        return True

    ### Expressions

    def _handle_expr_Call(self, expr: ailment.statement.Call) -> DataType:
        results = self._do_call(expr, True)
        if len(results) != 1:
            raise errors.AngrRuntimeError(f"Call expression returned with {len(results)} return values, expected 1")
        # these may be provided by misbehaving simprocedures. truncate the result as needed.
        result = results[0]
        if len(result) > expr.bits:
            assert isinstance(result, claripy.ast.BV)
            result = result[expr.bits - 1 : 0]
        elif len(result) < expr.bits:
            raise errors.AngrRuntimeError("Function returned too-small value for lvalue expression")
        return result

    def _handle_expr_Const(self, expr: ailment.expression.Const) -> claripy.ast.Bits:
        if isinstance(expr.value, int):
            return claripy.BVV(expr.value, expr.bits)
        if isinstance(expr.value, float):
            if expr.bits == 32:
                return claripy.FPV(expr.value, claripy.FSORT_FLOAT)
            if expr.bits == 64:
                return claripy.FPV(expr.value, claripy.FSORT_DOUBLE)
            raise TypeError(f"Bad width of float const {expr.bits}")
        raise TypeError(f"Bad const value type {type(expr.value)}")

    def _handle_expr_Register(self, expr: ailment.expression.Register) -> claripy.ast.Bits:
        return self.state.registers.load(expr.reg_offset, expr.size)

    def _handle_expr_Load(self, expr: ailment.expression.Load) -> claripy.ast.Bits:
        ptr = self._expr_bv(expr.addr)
        size = expr.size if isinstance(expr.size, int) else self._expr(expr.size)
        region, offset = self._find_ptr_region(ptr)
        return region.load(offset, size, endness=expr.endness)

    def _handle_expr_VirtualVariable(self, expr: ailment.expression.VirtualVariable) -> DataType:
        if isinstance((val := self.frame.vars.get(expr.varid, None)), MemoryMixin):
            return val.load(0, expr.size, endness=self.state.arch.memory_endness)
        if val is None:
            # unconstrained
            if angr.options.ZERO_FILL_UNCONSTRAINED_MEMORY in self.state.options:
                val = claripy.BVV(0, expr.bits)
            elif angr.options.SYMBOL_FILL_UNCONSTRAINED_MEMORY in self.state.options:
                val = claripy.BVS(f"unconstrained_vvar_{expr.varid}", expr.bits)
            else:
                raise errors.AngrRuntimeError(f"Load of vvar_{expr.varid} but it has no assigned value")
            self.frame.vars[expr.varid] = val
        return val

    def _handle_expr_Tmp(self, expr: ailment.expression.Tmp) -> DataType:
        return self.tmps[expr.tmp_idx]

    def _handle_expr_Phi(self, expr: ailment.expression.Phi) -> DataType:
        assert self.state.history.parent is not None
        last_addr = self.state.history.parent.recent_bbl_addrs[-1]
        for src, vvar in expr.src_and_vvars:
            if src == last_addr:
                assert vvar is not None
                return self._handle_expr_VirtualVariable(vvar)
        raise errors.AngrRuntimeError("None of the predecessors in a phi node are my predecessor!")

    def _handle_expr_Convert(self, expr: ailment.expression.Convert) -> DataType:
        child = self._expr(expr.operand)
        assert len(child) == expr.from_bits
        if expr.from_type == expr.TYPE_INT:
            if isinstance(child, claripy.ast.Bool):
                assert expr.from_bits == 1
                assert expr.to_type == expr.TYPE_INT
                return claripy.If(child, claripy.BVV(1, expr.to_bits), claripy.BVV(0, expr.to_bits))
            assert isinstance(child, claripy.ast.BV)
            if expr.to_type == expr.TYPE_INT:
                if expr.to_bits > expr.from_bits:
                    if expr.is_signed:
                        return child.sign_extend(expr.to_bits - expr.from_bits)
                    return child.zero_extend(expr.to_bits - expr.from_bits)
                return child[expr.to_bits - 1 : 0]
            if expr.to_type == expr.TYPE_FP:
                to_sort = claripy.FSORT_DOUBLE if expr.to_bits == 64 else claripy.FSORT_FLOAT
                return child.val_to_fp(to_sort, expr.is_signed, expr.rounding_mode)
            assert False
        elif expr.from_type == expr.TYPE_FP:
            assert isinstance(child, claripy.ast.FP)
            if expr.to_type == expr.TYPE_INT:
                return child.val_to_bv(expr.to_bits, expr.is_signed)
            if expr.to_type == expr.TYPE_FP:
                to_sort = claripy.FSORT_DOUBLE if expr.to_bits == 64 else claripy.FSORT_FLOAT
                return child.to_fp(to_sort, expr.rounding_mode)
            assert False
        else:
            assert False

    def _handle_expr_Reinterpret(self, expr: ailment.expression.Reinterpret) -> DataType:
        child = self._expr_bits(expr.operand)
        assert len(child) == expr.from_bits
        if expr.from_type == expr.TYPE_INT:
            assert isinstance(child, claripy.ast.BV)
            if expr.to_type == expr.TYPE_INT:
                assert False, "I think this is unreachable"
            elif expr.to_type == expr.TYPE_FP:
                assert expr.from_size == expr.to_size
                return child.raw_to_fp()
            else:
                assert False
        elif expr.from_type == expr.TYPE_FP:
            assert isinstance(child, claripy.ast.FP)
            if expr.to_type == expr.TYPE_INT:
                assert expr.from_size == expr.to_size
                return child.raw_to_bv()
            if expr.to_type == expr.TYPE_FP:
                assert False, "I think this is unreachable"
            else:
                assert False
        else:
            assert False

    def _handle_expr_ITE(self, expr: ailment.expression.ITE) -> DataType:
        cond = self._expr_bool(expr.cond)
        if cond.is_true():
            return self._expr_bits(expr.iftrue)
        if cond.is_false():
            return self._expr_bits(expr.iffalse)
        return claripy.If(cond, self._expr_bits(expr.iftrue), self._expr_bits(expr.iffalse))

    def _handle_expr_DirtyExpression(self, expr: ailment.expression.DirtyExpression) -> DataType:
        raise NotImplementedError(ailment.expression.DirtyExpression)

    def _handle_expr_VEXCCallExpression(self, expr: ailment.expression.VEXCCallExpression) -> DataType:
        handler = getattr(ccall, expr.callee, None)
        if handler is None:
            return self._top(expr.bits)
        args = tuple(self._expr_bits(arg) for arg in expr.operands)
        return handler(self.state, *args)

    def _handle_expr_MultiStatementExpression(self, expr: ailment.expression.MultiStatementExpression) -> DataType:
        for stmt in expr.stmts:
            self._stmt(stmt, False)
        return self._expr(expr.expr)

    def _handle_expr_BasePointerOffset(self, expr: ailment.expression.BasePointerOffset) -> DataType:
        return self._expr_bv(expr.base) + expr.offset

    def _handle_expr_StackBaseOffset(self, expr: ailment.expression.StackBaseOffset) -> DataType:
        assert self.frame.stack_ptr is not None
        return claripy.BVV(self.frame.stack_ptr + expr.offset, expr.bits)

    def _handle_unop_Neg(self, expr: ailment.UnaryOp):
        v = self._expr_bv(expr.operand)
        return -v

    def _handle_unop_Not(self, expr: ailment.UnaryOp):
        v = self._expr_bv(expr.operand)
        return ~v

    def _handle_unop_BitwiseNeg(self, expr: ailment.UnaryOp):
        v = self._expr_bv(expr.operand)
        return ~v

    def _handle_unop_Reference(self, expr: ailment.expression.UnaryOp) -> DataType:
        match expr.operand:
            case ailment.expression.VirtualVariable():
                curval = self.frame.vars.get(expr.operand.varid, None)
                if isinstance(curval, MemoryMixin):
                    return self.frame.var_refs_rev[expr.operand.varid]

                func_name = self.lift_addr(self.state.addr).function.name
                region_name = f"ail_engine_var_{func_name}_{expr.operand.varid}"

                # pick a class, any class...
                memory_cls = self.state.globals["ail_var_memory_cls"]  # type: ignore
                newval = memory_cls(memory_id=region_name)
                assert isinstance(newval, MemoryMixin)
                newval.set_state(self.state)
                if curval is not None:
                    newval.store(0, curval, endness=self.state.arch.memory_endness)
                newptr = claripy.BVS(region_name, expr.bits)
                self.frame.vars[expr.operand.varid] = newval
                self.frame.var_refs[newptr] = expr.operand.varid
                self.frame.var_refs_rev[expr.operand.varid] = newptr
                return newptr
            case _:
                raise errors.AngrRuntimeError(f"Can't handle reference to {expr.operand}")

    def _handle_unop_Dereference(self, expr: ailment.expression.UnaryOp) -> DataType:
        ptr = self._expr_bv(expr.operand)
        region, offset = self._find_ptr_region(ptr)
        return region.load(offset, expr.size, endness=expr.endness)

    def _handle_unop_Clz(self, expr: ailment.expression.UnaryOp) -> DataType:
        operand = self._expr(expr.operand)
        wtf_expr = claripy.BVV(expr.bits, expr.bits)
        for a in range(expr.bits):
            bit = claripy.Extract(a, a, operand)
            wtf_expr = claripy.If(bit == 1, claripy.BVV(expr.bits - a - 1, expr.bits), wtf_expr)
        return wtf_expr

    def _handle_unop_Ctz(self, expr: ailment.expression.UnaryOp) -> DataType:
        operand = self._expr(expr.operand)
        wtf_expr = claripy.BVV(expr.bits, expr.bits)
        for a in reversed(range(expr.bits)):
            bit = claripy.Extract(a, a, operand)
            wtf_expr = claripy.If(bit == 1, claripy.BVV(a, expr.bits), wtf_expr)
        return wtf_expr

    def _handle_unop_GetMSBs(self, expr: ailment.expression.UnaryOp) -> DataType:
        operand = self._expr(expr.operand)
        size = expr.bits
        bits = [claripy.Extract(i, i, operand) for i in range(size - 1, 6, -8)]
        return claripy.Concat(*bits)

    def _handle_unop_unpack(self, expr: ailment.expression.UnaryOp) -> DataType:
        raise NotImplementedError("Not sure what the semantics of this op are")

    def _handle_unop_Sqrt(self, expr: ailment.expression.UnaryOp) -> DataType:
        raise NotImplementedError("Not sure how to express this in claripy")

    def _handle_unop_RSqrtEst(self, expr: ailment.expression.UnaryOp) -> DataType:
        raise NotImplementedError("Not sure what the semantics of this op are")

    def _handle_binop_Add(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) + self._expr_bv(expr.operands[1])

    def _handle_binop_Sub(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) - self._expr_bv(expr.operands[1])

    def _handle_binop_Mul(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) * self._expr_bv(expr.operands[1])

    def _handle_binop_Div(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) / self._expr_bv(expr.operands[1])

    def _handle_binop_Mod(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) % self._expr_bv(expr.operands[1])

    def _handle_binop_AddF(self, expr: ailment.expression.BinaryOp) -> DataType:
        return claripy.ast.fp.fpAdd(
            expr.rounding_mode, self._expr_fp(expr.operands[0]), self._expr_fp(expr.operands[1])
        )

    def _handle_binop_SubF(self, expr: ailment.expression.BinaryOp) -> DataType:
        return claripy.ast.fp.fpSub(
            expr.rounding_mode, self._expr_fp(expr.operands[0]), self._expr_fp(expr.operands[1])
        )

    def _handle_binop_MulF(self, expr: ailment.expression.BinaryOp) -> DataType:
        return claripy.ast.fp.fpMul(
            expr.rounding_mode, self._expr_fp(expr.operands[0]), self._expr_fp(expr.operands[1])
        )

    def _handle_binop_DivF(self, expr: ailment.expression.BinaryOp) -> DataType:
        return claripy.ast.fp.fpDiv(
            expr.rounding_mode, self._expr_fp(expr.operands[0]), self._expr_fp(expr.operands[1])
        )

    def _handle_binop_AddV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        return claripy.Concat(
            *(
                a + b
                for a, b in zip(
                    self._expr_bv(expr.operands[0]).chop(expr.vector_size),
                    self._expr_bv(expr.operands[1]).chop(expr.vector_size),
                )
            )
        )

    def _handle_binop_SubV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        return claripy.Concat(
            *(
                a - b
                for a, b in zip(
                    self._expr_bv(expr.operands[0]).chop(expr.vector_size),
                    self._expr_bv(expr.operands[1]).chop(expr.vector_size),
                )
            )
        )

    def _handle_binop_MulV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        return claripy.Concat(
            *(
                a * b
                for a, b in zip(
                    self._expr_bv(expr.operands[0]).chop(expr.vector_size),
                    self._expr_bv(expr.operands[1]).chop(expr.vector_size),
                )
            )
        )

    def _handle_binop_DivV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        return claripy.Concat(
            *(
                a / b
                for a, b in zip(
                    self._expr_bv(expr.operands[0]).chop(expr.vector_size),
                    self._expr_bv(expr.operands[1]).chop(expr.vector_size),
                )
            )
        )

    def _handle_binop_Mull(self, expr: ailment.expression.BinaryOp) -> DataType:
        a = self._expr_bv(expr.operands[0])
        b = self._expr_bv(expr.operands[1])
        op = claripy.SignExt if expr.signed else claripy.ZeroExt
        return op(expr.bits - len(a), a) * op(expr.bits - len(b), b)

    def _handle_binop_MulHiV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure what the semantics of this op are")

    def _handle_binop_Xor(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) ^ self._expr_bv(expr.operands[1])

    def _handle_binop_And(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) & self._expr_bv(expr.operands[1])

    def _handle_binop_Or(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) | self._expr_bv(expr.operands[1])

    def _handle_binop_LogicalAnd(self, expr: ailment.expression.BinaryOp) -> DataType:
        a = self._expr_bv(expr.operands[0])
        if (a == 0).is_true():
            return a
        return claripy.If(a != 0, a, self._expr_bv(expr.operands[1]))

    def _handle_binop_LogicalOr(self, expr: ailment.expression.BinaryOp) -> DataType:
        a = self._expr_bv(expr.operands[0])
        if (a == 0).is_false():
            return a
        return claripy.If(a == 0, a, self._expr_bv(expr.operands[1]))

    def _handle_binop_Shl(self, expr: ailment.expression.BinaryOp) -> DataType:
        shift_amount = self._expr_bv(expr.operands[1])
        if shift_amount.size() < expr.bits:
            shift_amount = claripy.ZeroExt(expr.bits - shift_amount.size(), shift_amount)
        return self._expr_bv(expr.operands[0]) << shift_amount

    def _handle_binop_Shr(self, expr: ailment.expression.BinaryOp) -> DataType:
        shift_amount = self._expr_bv(expr.operands[1])
        if shift_amount.size() < expr.bits:
            shift_amount = claripy.ZeroExt(expr.bits - shift_amount.size(), shift_amount)
        return claripy.LShR(self._expr_bv(expr.operands[0]), shift_amount)

    def _handle_binop_Sar(self, expr: ailment.expression.BinaryOp) -> DataType:
        shift_amount = self._expr_bv(expr.operands[1])
        if shift_amount.size() < expr.bits:
            shift_amount = claripy.ZeroExt(expr.bits - shift_amount.size(), shift_amount)
        return self._expr_bv(expr.operands[0]) >> shift_amount

    def _handle_binop_CmpF(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure what the semantics of this op are")

    def _handle_binop_CmpEQ(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) == self._expr_bv(expr.operands[1])

    def _handle_binop_CmpNE(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) != self._expr_bv(expr.operands[1])

    def _handle_binop_CmpLT(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) < self._expr_bv(expr.operands[1])

    def _handle_binop_CmpLE(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) <= self._expr_bv(expr.operands[1])

    def _handle_binop_CmpGT(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) > self._expr_bv(expr.operands[1])

    def _handle_binop_CmpGE(self, expr: ailment.expression.BinaryOp) -> DataType:
        return self._expr_bv(expr.operands[0]) >= self._expr_bv(expr.operands[1])

    def _handle_binop_Concat(self, expr: ailment.expression.BinaryOp) -> DataType:
        return claripy.Concat(self._expr_bv(expr.operands[0]), self._expr_bv(expr.operands[1]))

    def _handle_binop_Rol(self, expr: ailment.expression.BinaryOp) -> DataType:
        shift_amount = self._expr_bv(expr.operands[1])
        if shift_amount.size() < expr.bits:
            shift_amount = claripy.ZeroExt(expr.bits - shift_amount.size(), shift_amount)
        return claripy.RotateLeft(self._expr_bv(expr.operands[0]), shift_amount)

    def _handle_binop_Ror(self, expr: ailment.expression.BinaryOp) -> DataType:
        shift_amount = self._expr_bv(expr.operands[1])
        if shift_amount.size() < expr.bits:
            shift_amount = claripy.ZeroExt(expr.bits - shift_amount.size(), shift_amount)
        return claripy.RotateRight(self._expr_bv(expr.operands[0]), shift_amount)

    def _handle_binop_Carry(self, expr: ailment.expression.BinaryOp) -> DataType:
        a = self._expr_bv(expr.operands[0])
        b = self._expr_bv(expr.operands[1])
        res = a + b
        return claripy.If(claripy.ULT(res, a), claripy.BVV(1, 1), claripy.BVV(0, 1))

    def _handle_binop_SCarry(self, expr: ailment.expression.BinaryOp) -> DataType:
        # TODO am I wrong here? why is this not just expr.signed??
        a = self._expr_bv(expr.operands[0])
        b = self._expr_bv(expr.operands[1])
        res = a + b
        return claripy.If(claripy.SLT(res, a), claripy.BVV(1, 1), claripy.BVV(0, 1))

    def _handle_binop_SBorrow(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_InterleaveLOV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        assert expr.vector_count is not None
        left_vector = [
            self._expr_bv(expr.operands[0])[(i + 1) * expr.vector_size - 1 : i * expr.vector_size]
            for i in range(expr.vector_count // 2)
        ]
        right_vector = [
            self._expr_bv(expr.operands[1])[(i + 1) * expr.vector_size - 1 : i * expr.vector_size]
            for i in range(expr.vector_count // 2)
        ]
        return claripy.Concat(*itertools.chain.from_iterable(zip(reversed(left_vector), reversed(right_vector))))

    def _handle_binop_InterleaveHIV(self, expr: ailment.expression.BinaryOp) -> DataType:
        assert expr.vector_size is not None
        assert expr.vector_count is not None
        left_vector = [
            self._expr_bv(expr.operands[0])[(i + 1) * expr.vector_size - 1 : i * expr.vector_size]
            for i in range(expr.vector_count // 2, expr.vector_count)
        ]
        right_vector = [
            self._expr_bv(expr.operands[1])[(i + 1) * expr.vector_size - 1 : i * expr.vector_size]
            for i in range(expr.vector_count // 2, expr.vector_count)
        ]
        return claripy.Concat(*itertools.chain.from_iterable(zip(reversed(left_vector), reversed(right_vector))))

    def _handle_binop_CasCmpEQ(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CasCmpNE(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_ExpCmpNE(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_SarN(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_SarNV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_ShrNV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_ShlNV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpEQV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpNEV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpGEV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpGTV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpLEV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpLTV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_MinV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_MaxV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_QAddV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_QNarrowBin(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_QNarrowBinV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_PermV(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_Set(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")

    def _handle_binop_CmpORD(self, expr: ailment.expression.BinaryOp) -> DataType:
        raise NotImplementedError("Not sure of the semantics of this op")



from __future__ import annotations

from .dream import DreamStructurer
from .phoenix import PhoenixStructurer
from .sailr import SAILRStructurer
from .recursive_structurer import RecursiveStructurer


STRUCTURER_CLASSES = {
    SAILRStructurer.NAME: SAILRStructurer,
    PhoenixStructurer.NAME: PhoenixStructurer,
    DreamStructurer.NAME: DreamStructurer,
}

DEFAULT_STRUCTURER = SAILRStructurer


def structurer_class_from_name(name: str) -> type | None:
    return STRUCTURER_CLASSES.get(name.lower(), None)


__all__ = (
    "DEFAULT_STRUCTURER",
    "STRUCTURER_CLASSES",
    "DreamStructurer",
    "PhoenixStructurer",
    "RecursiveStructurer",
    "SAILRStructurer",
    "structurer_class_from_name",
)



from __future__ import annotations
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary

libc = SimLibrary()
libc.set_library_names(
    "libSystem.dylib",
    "libSystem.B.dylib",
)
# TODO: add more functions
libc.add_all_from_dict(P["libc"])

# Mach-O naming convention adds an underscore prefix to function names
for name, _ in P["libc"].items():
    libc.add_alias(name, "_" + name)



# pylint:disable=multiple-statements,line-too-long,consider-using-enumerate
from __future__ import annotations
from typing import Any, TYPE_CHECKING
import logging
from collections import defaultdict, OrderedDict

import networkx

import claripy
import angr.ailment as ailment

from angr.utils.graph import GraphUtils
from angr.knowledge_plugins.cfg import IndirectJumpType
from angr.analyses.decompiler.graph_region import GraphRegion
from angr.analyses.decompiler.empty_node_remover import EmptyNodeRemover
from angr.analyses.decompiler.jumptable_entry_condition_rewriter import JumpTableEntryConditionRewriter
from angr.analyses.decompiler.condition_processor import ConditionProcessor
from angr.analyses.decompiler.region_simplifiers.cascading_cond_transformer import CascadingConditionTransformer
from angr.analyses.decompiler.utils import (
    extract_jump_targets,
    get_ast_subexprs,
    switch_extract_cmp_bounds,
    remove_last_statement,
    first_nonlabel_nonphi_node,
)
from .structurer_nodes import (
    SequenceNode,
    CodeNode,
    ConditionNode,
    ConditionalBreakNode,
    LoopNode,
    SwitchCaseNode,
    BreakNode,
    ContinueNode,
    MultiNode,
    CascadingConditionNode,
    BaseNode,
    EmptyBlockNotice,
)
from .structurer_base import StructurerBase


if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function

l = logging.getLogger(name=__name__)


#
# The main analysis
#


class DreamStructurer(StructurerBase):
    """
    Structure a region using a structuring algorithm that is similar to the one in Dream decompiler (described in the
    "no more gotos" paper). Note that this implementation has quite a few improvements over the original described
    version and *should not* be used to evaluate the performance of the original algorithm described in that paper.

    The current function graph is provided so that we can detect certain edge cases, for example, jump table entries no
    longer exist due to empty node removal during structuring or prior steps.
    """

    NAME = "dream"

    def __init__(
        self,
        region,
        parent_map=None,
        condition_processor=None,
        func: Function | None = None,
        case_entry_to_switch_head: dict[int, int] | None = None,
        parent_region=None,
        **kwargs,
    ):
        super().__init__(
            region,
            parent_map=parent_map,
            condition_processor=condition_processor,
            func=func,
            case_entry_to_switch_head=case_entry_to_switch_head,
            parent_region=parent_region,
            **kwargs,
        )

        self._analyze()

    def _analyze(self):
        has_cycle = self._has_cycle()
        # sanity checks
        if self._region.cyclic:
            if not has_cycle:
                l.critical(
                    "Region %r is supposed to be a cyclic region but there is no cycle inside. This is usually "
                    "due to the existence of loop headers with more than one in-edges, which angr decompiler "
                    "does not support yet. The decompilation result will be wrong.",
                    self._region,
                )
            self._analyze_cyclic()
        else:
            if has_cycle:
                l.critical(
                    "Region %r is supposed to be an acyclic region but there are cycles inside. This is usually "
                    "due to the existence of loop headers with more than one in-edges, which angr decompiler "
                    "does not support yet. The decompilation result will be wrong.",
                    self._region,
                )
            self._analyze_acyclic()

    def _analyze_cyclic(self):
        loop_head = self._region.head

        loop_subgraph = self._region.graph
        successors = self._region.successors
        assert successors is not None

        assert len(successors) <= 1

        loop_node = self._make_endless_loop(loop_head, loop_subgraph, successors)

        loop_node = self._refine_loop(loop_node)

        seq = SequenceNode(
            loop_head.addr, nodes=[loop_node] + [succ for succ in successors if succ in self._region.graph]
        )

        self.result = seq

    def _analyze_acyclic(self):
        # let's generate conditions first
        self.cond_proc.recover_reaching_conditions(
            self._region, with_successors=True, case_entry_to_switch_head=self._case_entry_to_switch_head
        )

        # make the sequence node and pack reaching conditions into CodeNode instances
        seq = self._make_sequence()

        self._new_sequences.append(seq)

        while self._new_sequences:
            seq_ = self._new_sequences.pop(0)
            if len(seq_.nodes) <= 1:
                continue
            self._structure_sequence(seq_)

        seq = EmptyNodeRemover(seq).result

        # unpack nodes and remove CodeNode wrappers
        seq = self._unpack_sequence(seq)

        self.result = seq

    def _find_loop_nodes_and_successors(self):
        graph = self._region.graph
        head = self._region.head

        # find initial loop nodes
        loop_nodes = None
        components = networkx.strongly_connected_components(graph)
        for component in components:
            if head in component:
                loop_nodes = component
                break
        if loop_nodes is None:
            # this should never happen - loop head always forms a cycle
            raise TypeError("A bug (impossible case) in the algorithm is triggered.")

        # extend loop nodes
        while True:
            loop_nodes_updated = False
            for loop_node in loop_nodes:
                for succ in graph.successors(loop_node):
                    if succ not in loop_nodes:
                        # determine if this successor's all predecessors are in the loop
                        predecessors = graph.predecessors(succ)
                        if all(pred in loop_nodes for pred in predecessors):
                            # yes!
                            loop_nodes.add(succ)
                            loop_nodes_updated = True
                            break
                if loop_nodes_updated:
                    break
            if not loop_nodes_updated:
                break

        # find loop nodes and successors
        loop_subgraph = networkx.subgraph(graph, loop_nodes)
        loop_node_addrs = {node.addr for node in loop_subgraph}

        # Case A: The loop successor is inside the current region (does it happen at all?)
        loop_successors = set()

        for node, successors in networkx.bfs_successors(graph, head):
            if node.addr in loop_node_addrs:
                for suc in successors:
                    if suc not in loop_subgraph:
                        loop_successors.add(suc)

        # Case B: The loop successor is the successor to this region in the parent graph
        if not loop_successors and self._parent_map is not None:
            current_region = self._region
            parent_region = self._parent_map.get(current_region, None)
            while parent_region and not loop_successors:
                parent_graph = parent_region.graph
                for node, successors in networkx.bfs_successors(parent_graph, current_region):
                    if node.addr == current_region.addr:
                        for suc in successors:
                            if suc not in loop_subgraph:
                                loop_successors.add(suc)
                current_region = parent_region
                parent_region = self._parent_map.get(current_region, None)

        return loop_subgraph, loop_successors

    def _make_endless_loop(self, loop_head, loop_subgraph, loop_successors):
        loop_body = self._to_loop_body_sequence(loop_head, loop_subgraph, loop_successors)

        # create a while(true) loop with sequence node being the loop body
        return LoopNode("while", None, loop_body, addr=loop_head.addr)

    def _refine_loop(self, loop_node):
        while True:
            # while
            r, loop_node = self._refine_loop_while(loop_node)
            if r:
                continue

            # do-while
            r, loop_node = self._refine_loop_dowhile(loop_node)
            if r:
                continue

            # no more changes
            break

        return loop_node

    @staticmethod
    def _refine_loop_while(loop_node):
        if loop_node.sort == "while" and loop_node.condition is None and loop_node.sequence_node.nodes:
            # it's an endless loop
            first_node = first_nonlabel_nonphi_node(loop_node.sequence_node)
            inner_first_node = first_node.node if type(first_node) is CodeNode else first_node
            if type(inner_first_node) is ConditionalBreakNode:
                while_cond = ConditionProcessor.simplify_condition(claripy.Not(inner_first_node.condition))
                new_seq = loop_node.sequence_node.copy()
                new_seq.nodes = [nn for nn in new_seq.nodes if nn is not first_node]
                new_loop_node = LoopNode("while", while_cond, new_seq, addr=loop_node.addr)

                return True, new_loop_node

        return False, loop_node

    @staticmethod
    def _refine_loop_dowhile(loop_node):
        if loop_node.sort == "while" and loop_node.condition is None and loop_node.sequence_node.nodes:
            # it's an endless loop
            last_node = loop_node.sequence_node.nodes[-1]
            if type(last_node) is ConditionalBreakNode:
                while_cond = ConditionProcessor.simplify_condition(claripy.Not(last_node.condition))
                new_seq = loop_node.sequence_node.copy()
                new_seq.nodes = new_seq.nodes[:-1]
                new_loop_node = LoopNode("do-while", while_cond, new_seq)

                return True, new_loop_node

        return False, loop_node

    def _to_loop_body_sequence(self, loop_head, loop_subgraph, loop_successors):
        graph = self._region.graph_with_successors
        assert graph is not None
        loop_region_graph = networkx.DiGraph()

        # TODO: Make sure the loop body has been structured

        queue = [loop_head]
        traversed = set()
        loop_successor_addrs = {succ.addr for succ in loop_successors}
        replaced_nodes = {}
        outedges = []

        while queue:
            node = queue[0]
            queue = queue[1:]

            loop_region_graph.add_node(node)
            traversed.add(node)

            successors_and_data = list(graph.out_edges(node, data=True))  # successors are all inside the current region

            for _, dst, edge_data in successors_and_data:
                # sanity check
                if dst.addr in loop_successor_addrs:
                    outedges.append((node, dst, edge_data))
                    continue
                if dst not in loop_subgraph and dst.addr not in loop_successor_addrs:
                    # what's this node?
                    l.error("Found a node that belongs to neither loop body nor loop successors. Something is wrong.")
                    # raise Exception()

                if replaced_nodes.get(dst, dst) is not loop_head:
                    loop_region_graph.add_edge(node, replaced_nodes.get(dst, dst), **edge_data)
                if dst in traversed or dst in queue:
                    continue
                queue.append(dst)

        # Create a graph region and structure it
        loop_region_graph_with_successors = networkx.DiGraph(loop_region_graph)
        loop_successors = set()  # update loop_successors with nodes in outedges
        for src, dst, edge_data in outedges:
            loop_region_graph_with_successors.add_edge(src, dst, **edge_data)
            loop_successors.add(dst)
        region = GraphRegion(
            loop_head, loop_region_graph, successors=None, graph_with_successors=None, cyclic=False, full_graph=None
        )
        structurer = self.project.analyses[DreamStructurer].prep()(
            region, condition_processor=self.cond_proc, func=self.function, jump_tables=self.jump_tables
        )
        seq = structurer.result

        # traverse this node and rewrite all conditional jumps that go outside the loop to breaks
        self._rewrite_conditional_jumps_to_breaks(seq, loop_successor_addrs)
        # traverse this node and rewrite all jumps that go to the beginning of the loop to continue
        self._rewrite_jumps_to_continues(seq)

        seq = self._remove_redundant_jumps(seq)
        seq = self._remove_conditional_jumps(seq)
        seq = EmptyNodeRemover(seq).result

        while True:
            r, seq = self._merge_conditional_breaks(seq)
            if r:
                continue
            r, seq = self._merge_nesting_conditionals(seq)
            if r:
                continue
            break

        return EmptyNodeRemover(seq).result

    def _make_sequence(self):
        seq = SequenceNode(None)

        for node in GraphUtils.quasi_topological_sort_nodes(self._region.graph):
            seq.add_node(CodeNode(node, self.cond_proc.reaching_conditions.get(node, None)))

        if seq.nodes:
            seq.addr = seq.nodes[0].addr

        return seq

    @staticmethod
    def _unpack_sequence(seq):
        def _handle_Code(node, **kwargs):  # pylint:disable=unused-argument
            node = node.node
            return walker._handle(node)

        def _handle_Sequence(node, **kwargs):  # pylint:disable=unused-argument
            for i in range(len(node.nodes)):  # pylint:disable=consider-using-enumerate
                node.nodes[i] = walker._handle(node.nodes[i])
            return node

        def _handle_ConditionNode(node, **kwargs):  # pylint:disable=unused-argument
            if node.true_node is not None:
                node.true_node = walker._handle(node.true_node)
            if node.false_node is not None:
                node.false_node = walker._handle(node.false_node)
            return node

        def _handle_CascadingConditionNode(node: CascadingConditionNode, **kwargs):  # pylint:disable=unused-argument
            new_cond_and_nodes = []
            for cond, child_node in node.condition_and_nodes:
                new_cond_and_nodes.append((cond, walker._handle(child_node)))
            node.condition_and_nodes = new_cond_and_nodes

            if node.else_node is not None:
                node.else_node = walker._handle(node.else_node)  # type: ignore
            return node

        def _handle_SwitchCaseNode(node, **kwargs):  # pylint:disable=unused-argument
            for i in list(node.cases.keys()):
                node.cases[i] = walker._handle(node.cases[i])
            if node.default_node is not None:
                node.default_node = walker._handle(node.default_node)
            return node

        def _handle_Default(node, **kwargs):  # pylint:disable=unused-argument
            return node

        handlers = {
            CodeNode: _handle_Code,
            SequenceNode: _handle_Sequence,
            ConditionNode: _handle_ConditionNode,
            CascadingConditionNode: _handle_CascadingConditionNode,
            SwitchCaseNode: _handle_SwitchCaseNode,
            # don't do anything
            LoopNode: _handle_Default,
            ContinueNode: _handle_Default,
            ConditionalBreakNode: _handle_Default,
            BreakNode: _handle_Default,
            MultiNode: _handle_Default,
            ailment.Block: _handle_Default,
        }

        walker = SequenceWalker(handlers=handlers)
        walker.walk(seq)

        return seq

    def _structure_sequence(self, seq):
        self._make_switch_cases(seq)

        # this is hackish...
        # seq.nodes = new_seq.nodes

        self._merge_same_conditioned_nodes(seq)
        self._structure_common_subexpression_conditions(seq)
        self._make_ites(seq)
        self._remove_redundant_jumps(seq)

        empty_node_remover = EmptyNodeRemover(seq)
        new_seq = empty_node_remover.result
        # update self._new_sequences
        self._update_new_sequences(set(empty_node_remover.removed_sequences), empty_node_remover.replaced_sequences)

        # we need to do it in-place
        seq.nodes = new_seq.nodes

        self._replace_complex_reaching_conditions(seq)
        self._make_condition_nodes(seq)
        self._make_cascading_condition_nodes(seq)

        while True:
            r, seq = self._merge_conditional_breaks(seq)
            if r:
                continue
            r, seq = self._merge_nesting_conditionals(seq)
            if r:
                continue
            break

    def _merge_same_conditioned_nodes(self, seq):
        # search for nodes with the same reaching condition and then merge them into one sequence node
        i = 0
        while i < len(seq.nodes) - 1:
            node_0 = seq.nodes[i]
            if type(node_0) is not CodeNode:
                i += 1
                continue
            rcond_0 = node_0.reaching_condition
            if rcond_0 is None:
                i += 1
                continue
            node_1 = seq.nodes[i + 1]
            if type(node_1) is not CodeNode:
                i += 1
                continue
            rcond_1 = node_1.reaching_condition
            if rcond_1 is None:
                i += 1
                continue
            r = claripy.simplify(rcond_0 == rcond_1)
            if claripy.is_true(r):
                # node_0 and node_1 should be put into the same sequence node
                new_node = CodeNode(
                    self._merge_nodes(node_0.node, node_1.node),
                    node_0.reaching_condition,
                )
                seq.nodes = [*seq.nodes[:i], new_node, *seq.nodes[i + 2 :]]
                continue
            i += 1

    #
    # Dealing with switch-case structures
    #

    def _make_switch_cases(self, seq):
        """
        Search for nodes that look like switch-cases and convert them to switch cases.

        :param seq:     The Sequence node.
        :return:        None
        """

        addr2nodes: dict[int, set[CodeNode]] = defaultdict(set)
        for node in seq.nodes:
            addr2nodes[node.addr].add(node)

        while True:
            for i in range(len(seq.nodes)):
                node = seq.nodes[i]

                # Jumptable_AddressLoadedFromMemory
                r = self._make_switch_cases_address_loaded_from_memory(seq, i, node, addr2nodes)
                if r:
                    # we found a node that looks like a switch-case. seq.nodes are changed. resume to find the next such
                    # case
                    break

                # Jumptable_AddressComputed
                r = self._make_switch_cases_address_computed(seq, i, node, addr2nodes)
                if r:
                    break

            else:
                # we did not find any node that looks like a switch-case. exit.
                break

    def _make_switch_cases_address_loaded_from_memory(self, seq, i, node, addr2nodes: dict[int, set[CodeNode]]) -> bool:
        """
        A typical jump table involves multiple nodes, which look like the following:

        Head:  s_50 = Conv(32->64, (Load(addr=stack_base-28, size=4, endness=Iend_LE) - 0x3f<32>))<8>
               if (((Load(addr=stack_base-28, size=4, endness=Iend_LE) - 0x3f<32>) <= 0x36<32>))
                    { Goto A<64> } else { Goto B<64> }

        A:     (with an indirect jump)
               Goto((
                Conv(32->64, Load(addr=(0x40964c<64> + (Load(addr=stack_base-80, size=8, endness=Iend_LE) Mul 0x4<8>)),
                size=4, endness=Iend_LE)) + 0x40964c<64>))

        B:     (the default case)
        """

        try:
            last_stmt = self.cond_proc.get_last_statement(node)
            if not isinstance(last_stmt, ailment.Stmt.ConditionalJump):
                return False
        except EmptyBlockNotice:
            return False
        successor_addrs = extract_jump_targets(last_stmt)
        if len(successor_addrs) != 2:
            return False

        for t in successor_addrs:
            if t in addr2nodes and t in self.jump_tables:
                # this is a candidate!
                target = t
                break
        else:
            return False

        jump_table = self.jump_tables[target]
        if jump_table.type != IndirectJumpType.Jumptable_AddressLoadedFromMemory:
            return False

        # extract the comparison expression, lower-, and upper-bounds from the last statement
        cmp = switch_extract_cmp_bounds(last_stmt)
        if not cmp:
            return False
        cmp_expr, cmp_lb, _cmp_ub = cmp  # pylint:disable=unused-variable

        # the real indirect jump
        if len(addr2nodes[target]) != 1:
            return False
        node_a = next(iter(addr2nodes[target]))
        # the default case
        node_b_addr = next(iter(t for t in successor_addrs if t != target))

        # Node A might have been structured. Un-structure it if that is the case.
        r, node_a = self._switch_unpack_sequence_node(seq, node_a, node_b_addr, jump_table, addr2nodes)
        if not r:
            return False

        # build switch-cases
        assert jump_table.jumptable_entries is not None
        cases, node_default, to_remove = self._switch_build_cases(
            seq, cmp_lb, jump_table.jumptable_entries, i, node_b_addr, addr2nodes
        )
        # if we don't know what the end address of this switch-case structure is, let's figure it out
        switch_end_addr = (
            node_b_addr
            if node_default is None
            else self._switch_find_switch_end_addr(cases, node_default, {nn.addr for nn in self._region.graph})
        )
        if switch_end_addr is not None:
            self._switch_handle_gotos(cases, node_default, switch_end_addr)

        assert last_stmt.tags["ins_addr"] is not None
        self._make_switch_cases_core(
            seq,
            i,
            node,
            cmp_expr,
            cases,
            node_default,
            last_stmt.tags["ins_addr"],
            addr2nodes,
            to_remove,
            node_a=node_a,
            jumptable_addr=jump_table.addr,
        )

        return True

    def _make_switch_cases_address_computed(self, seq, i, node, addr2nodes: dict[int, set[CodeNode]]) -> bool:
        if node.addr not in self.jump_tables:
            return False
        jump_table = self.jump_tables[node.addr]
        if jump_table.type != IndirectJumpType.Jumptable_AddressComputed:
            return False

        try:
            last_stmts = self.cond_proc.get_last_statements(node)
        except EmptyBlockNotice:
            return False
        if len(last_stmts) != 1:
            return False
        last_stmt = last_stmts[0]

        if not isinstance(last_stmt, ailment.Stmt.ConditionalJump):
            return False

        # Typical look:
        #   t2 = (r5<4> - 0x22<32>)
        #   if ((t2 <= 0x1c<32>)) { Goto (0x41d10c<32> + (t2 << 0x2<8>)) } else { Goto 0x41d108<32> }
        #
        # extract the comparison expression, lower-, and upper-bounds from the last statement
        cmp = switch_extract_cmp_bounds(last_stmt)
        if not cmp:
            return False
        cmp_expr, cmp_lb, _cmp_ub = cmp  # pylint:disable=unused-variable

        jumptable_entries = jump_table.jumptable_entries
        assert jumptable_entries is not None

        if isinstance(last_stmt.false_target, ailment.Expr.Const):
            default_addr = last_stmt.false_target.value
            assert isinstance(default_addr, int)
        else:
            return False

        cases, node_default, to_remove = self._switch_build_cases(
            seq, cmp_lb, jumptable_entries, i, default_addr, addr2nodes
        )
        if node_default is None:
            # there must be a default case
            return False

        self._make_switch_cases_core(
            seq,
            i,
            node,
            cmp_expr,
            cases,
            node_default,
            node.addr,
            addr2nodes,
            to_remove,
            jumptable_addr=jump_table.addr,
        )

        return True

    def _make_switch_cases_core(
        self,
        seq,
        i,
        node,
        cmp_expr,
        cases: OrderedDict,
        node_default,
        addr,
        addr2nodes,
        to_remove,
        *,
        jumptable_addr: int,
        node_a=None,
    ):
        scnode = SwitchCaseNode(cmp_expr, cases, node_default, addr=addr)
        scnode = CodeNode(scnode, node.reaching_condition)

        # insert the switch-case node
        seq.insert_node(i + 1, scnode)
        # remove all those entry nodes
        if node_default is not None:
            to_remove.add(node_default)
        for node_ in to_remove:
            seq.remove_node(node_)
            addr2nodes[node_.addr].discard(node_)
            if not addr2nodes[node_.addr]:
                del addr2nodes[node_.addr]
        # remove the last statement in node
        remove_last_statement(node)
        if BaseNode.test_empty_node(node):
            seq.remove_node(node)
        if node_a is not None:
            # remove the last statement in node_a
            remove_last_statement(node_a)
            if BaseNode.test_empty_node(node_a):
                seq.remove_node(node_a)

        # rewrite conditions in the entire SequenceNode to remove jump table entry conditions
        rewriter = JumpTableEntryConditionRewriter(self.cond_proc.jump_table_conds[jumptable_addr])
        rewriter.walk(seq)  # update SequenceNodes in-place

    def _switch_unpack_sequence_node(
        self, seq: SequenceNode, node_a, node_b_addr: int, jumptable, addr2nodes: dict[int, set[CodeNode]]
    ) -> tuple[bool, CodeNode | None]:
        """
        We might have already structured the actual body of the switch-case structure into a single Sequence node (node
        A). If that is the case, we un-structure the sequence node in this method.

        :param seq:                 The original Sequence node.
        :param node_a:              Node A.
        :param node_b_addr:         Address of node B.
        :param jumptable:           The corresponding jump table instance.
        :param addr2nodes:          A dict of addresses to their corresponding nodes in `seq`.
        :return:                    A boolean value indicating the result and an updated node_a. The boolean value is
                                    True if unpacking is not necessary or we successfully unpacked the sequence node,
                                    False otherwise.
        """

        jumptable_entries = jumptable.jumptable_entries

        node_a_block_addrs = {n.addr for n in node_a.node.nodes} if isinstance(node_a.node, SequenceNode) else set()
        #
        # if that is the case, we un-structure it here
        if all(entry_addr in addr2nodes for entry_addr in jumptable_entries):
            return True, node_a
        if self._switch_check_existence_of_jumptable_entries(
            jumptable_entries, node_a_block_addrs, set(addr2nodes.keys()), node_a.addr, node_b_addr
        ):
            # unpacking is needed
            for n in node_a.node.nodes:
                assert n.addr is not None
                if isinstance(n, ConditionNode):
                    unpacked = self._switch_unpack_condition_node(n, jumptable)
                    if unpacked is None:
                        # unsupported. bail
                        return False, None
                    if n.addr in addr2nodes:
                        del addr2nodes[n.addr]
                    addr2nodes[n.addr].add(unpacked)
                    seq.add_node(unpacked)
                else:
                    the_node = CodeNode(n, None)
                    if n.addr in addr2nodes:
                        del addr2nodes[n.addr]
                    addr2nodes[n.addr].add(the_node)
                    seq.add_node(the_node)
            if node_a != addr2nodes[node_a.addr]:
                # update node_a
                seq.remove_node(node_a)
                node_a = next(iter(addr2nodes[node_a.addr]))
            return True, node_a

        # a jumptable entry is missing. it's very likely marked as the successor of the entire switch-case region. we
        # should have been handling it when dealing with multi-exit regions. ignore it here.
        return True, node_a

    def _switch_unpack_condition_node(self, cond_node: ConditionNode, jumptable) -> CodeNode | None:
        """
        Unpack condition nodes by only removing one condition in the form of
        <Bool jump_table_402020 == 0x402ac4>.

        :param cond_node:   The condition node to unpack.
        :return:            The new unpacked node.
        """

        # FIXME: With the new jump table entry condition, this function is probably never used. Remove sequence node
        # FIXME: unpacking logic if that is the case.

        cond = cond_node.condition

        # look for a condition in the form of xxx == jump_target
        eq_condition = None
        remaining_cond = None
        true_node = None
        false_node = None

        jumptable_var = self.cond_proc.create_jump_target_var(jumptable.addr)

        if cond.op == "And":
            for arg in cond.args:
                if (
                    arg.op == "__eq__"
                    and arg.args[0] is jumptable_var
                    and isinstance(arg.args[1], claripy.ast.Bits)
                    and arg.args[1].concrete
                ):
                    # found it
                    eq_condition = arg
                    remaining_cond = claripy.And(*(arg_ for arg_ in cond.args if arg_ is not arg))
                    true_node = cond_node.true_node
                    false_node = cond_node.false_node
                    break
            else:
                # unsupported
                return None
        elif cond.op == "__eq__":
            if cond.args[0] is jumptable_var and isinstance(cond.args[1], claripy.ast.Bits) and cond.args[1].concrete:
                # found it
                eq_condition = cond
                true_node = cond_node.true_node
                false_node = cond_node.false_node
                remaining_cond = None
            else:
                # unsupported
                return None
        else:
            # unsupported
            return None

        if remaining_cond is None:
            if true_node is not None and false_node is None:
                return CodeNode(true_node, eq_condition)
            # unsupported
            return None

        return CodeNode(
            ConditionNode(cond_node.addr, claripy.true(), remaining_cond, true_node, false_node=false_node),
            eq_condition,
        )

    def _switch_check_existence_of_jumptable_entries(
        self,
        jumptable_entries,
        node_a_block_addrs: set[int],
        known_node_addrs: set[int],
        node_a_addr: int,
        node_b_addr: int,
    ) -> bool:
        """
        Check if all entries in the given jump table exist in the given set of nodes of a SequenceNode.

        :param jumptable_entries:   Addresses of jump table entries.
        :param node_a_block_addrs:  A set of addresses for nodes that belong to Node A.
        :return:                    True if the check passes, False otherwise.
        """

        all_node_addrs = node_a_block_addrs | known_node_addrs | {node_b_addr}
        expected_node_a_addrs = set()
        for entry_addr in jumptable_entries:
            if entry_addr in all_node_addrs:
                expected_node_a_addrs.add(entry_addr)
                continue
            # the entry may go missing if the entire node has been folded into its successor node.
            # in this case, we check if (a) this entry node has only one successor, and (b) this successor exists in
            # seq_node_addrs.
            if self.function is not None:
                entry_node = self.function.get_node(entry_addr)
                if entry_node is not None:
                    successors = []
                    for _, dst, data in self.function.graph.out_edges(entry_node, data=True):
                        if data.get("type", "transition") != "call":
                            successors.append(dst)
                    if len(successors) == 1 and successors[0].addr in all_node_addrs:
                        # found the single successor
                        expected_node_a_addrs.add(successors[0].addr)
                        continue
            # it's also possible that this is just a jump that breaks out of the switch-case. we simply ignore it.
            continue

        # finally, make sure all expected nodes exist
        # not sure what is going on... if it's false
        return node_a_block_addrs.issuperset((expected_node_a_addrs | {node_a_addr}) - {node_b_addr})

    def _switch_find_jumptable_entry_node(self, entry_addr: int, addr2nodes: dict[int, set[CodeNode]]) -> Any | None:
        """
        Find the correct node for a given jump table entry address in addr2nodes.

        This method is needed because prior optimization steps may remove some blocks (e.g., empty blocks or blocks that
        only have branch instructions). If the given jump table entry address corresponds to a removed block, it will
        not be found inside addr2nodes dict. In such cases, we need to follow graph edges in the CFG and find the first
        block whose address is inside addr2nodes dict.

        :param entry_addr:  Address of the jump table entry.
        :return:            The correct node if we can find it, or None if we fail to find one.
        """

        if entry_addr in addr2nodes and len(addr2nodes[entry_addr]) == 1:
            return next(iter(addr2nodes[entry_addr]))
        # magic
        if self.function is None:
            return None

        addr = entry_addr
        node = self.function.get_node(addr)
        for _ in range(5):  # we try at most five steps
            if node is None:
                return None
            successors = []
            for _, dst, data in self.function.graph.out_edges(node, data=True):
                if data.get("type", "transition") != "call":
                    successors.append(dst)
            if len(successors) != 1:
                return None
            successor = successors[0]
            if successor.addr in addr2nodes:
                # found it!
                return next(iter(addr2nodes[successor.addr]))
            # keep looking
            node = successor
        return None

    def _switch_build_cases(
        self,
        seq: SequenceNode,
        cmp_lb: int,
        jumptable_entries: list[int],
        head_node_idx: int,
        node_b_addr: int,
        addr2nodes: dict[int, set[CodeNode]],
    ) -> tuple[OrderedDict, Any, Any]:
        """
        Discover all cases for the switch-case structure and build the switch-cases dict.

        :param seq:                 The original Sequence node.
        :param cmp_lb:              The lower bound of the jump table comparison.
        :param jumptable_entries:   Addresses of indirect jump targets in the jump table.
        :param head_node_addr:      The index of the head block of this jump table in `seq`.
        :param node_b_addr:         Address of node B. Potentially, node B is the default node.
        :param addr2nodes:          A dict of addresses to their corresponding nodes in `seq`.
        :return:                    A tuple of (dict of cases, the default node if exists, nodes to remove).
        """

        cases: OrderedDict[int | tuple[int, ...], SequenceNode] = OrderedDict()
        to_remove = set()
        node_default = addr2nodes.get(node_b_addr)
        if node_default is not None:
            node_default = next(iter(node_default))

        entry_addrs_set = set(jumptable_entries)
        converted_nodes: dict[int, Any] = {}
        entry_addr_to_ids = defaultdict(set)

        for j, entry_addr in enumerate(jumptable_entries):
            cases_idx = cmp_lb + j
            if entry_addr == node_b_addr:
                # jump to default or end of the switch-case structure - ignore this case
                continue

            entry_addr_to_ids[entry_addr].add(cases_idx)

            if entry_addr in converted_nodes:
                continue

            entry_node = self._switch_find_jumptable_entry_node(entry_addr, addr2nodes)
            if entry_node is None:
                # Missing entries. They are probably *after* the entire switch-case construct. Replace it with an empty
                # Goto node.
                case_inner_node = ailment.Block(
                    0,
                    0,
                    statements=[
                        ailment.Stmt.Jump(
                            None,
                            ailment.Expr.Const(None, None, entry_addr, self.project.arch.bits),
                            ins_addr=0,
                            stmt_idx=0,
                        )
                    ],
                )
                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true())])
                converted_nodes[entry_addr] = case_node
                continue

            case_node = SequenceNode(entry_node.addr, nodes=[CodeNode(entry_node.node, claripy.true())])
            to_remove.add(entry_node)
            entry_node_idx = seq.nodes.index(entry_node)

            if entry_node_idx <= head_node_idx:
                # it's jumping to a block that dominates the head. it's likely to be an optimized continue; statement
                # in a switch-case wrapped inside a while loop.
                # replace it with an empty Goto node
                case_inner_node = ailment.Block(
                    0,
                    0,
                    statements=[
                        ailment.Stmt.Jump(
                            None,
                            ailment.Expr.Const(None, None, entry_addr, self.project.arch.bits),
                            ins_addr=0,
                            stmt_idx=0,
                        )
                    ],
                )
                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true())])
                converted_nodes[entry_addr] = case_node
                continue

            # find nodes that this entry node dominates
            cond_subexprs = list(get_ast_subexprs(entry_node.reaching_condition))
            guarded_nodes = None
            for subexpr in cond_subexprs:
                guarded_node_candidates = self._nodes_guarded_by_common_subexpr(seq, subexpr, entry_node_idx + 1)
                if guarded_nodes is None:
                    guarded_nodes = {node_ for _, node_, _ in guarded_node_candidates}
                else:
                    guarded_nodes = guarded_nodes.intersection({node_ for _, node_, _ in guarded_node_candidates})

            if guarded_nodes is not None:
                # keep the topological order of nodes in Sequence node
                sorted_guarded_nodes = [node_ for node_ in seq.nodes[entry_node_idx + 1 :] if node_ in guarded_nodes]
                for node_ in sorted_guarded_nodes:
                    if node_ is not entry_node and node_.addr not in entry_addrs_set:
                        # fix reaching condition
                        reaching_condition_subexprs = set(get_ast_subexprs(node_.reaching_condition)).difference(
                            set(cond_subexprs)
                        )
                        new_reaching_condition = claripy.And(*reaching_condition_subexprs)
                        new_node = CodeNode(node_.node, new_reaching_condition)
                        case_node.add_node(new_node)
                        to_remove.add(node_)

            # do we have a default node?
            case_last_stmt = self.cond_proc.get_last_statement(case_node)
            if isinstance(case_last_stmt, ailment.Stmt.Jump):
                targets = extract_jump_targets(case_last_stmt)
                if len(targets) == 1 and targets[0] == node_b_addr:
                    # jump to the default case is rare - it's more likely that there is no default for this
                    # switch-case struct
                    node_default = None

            converted_nodes[entry_addr] = case_node

        for entry_addr, converted_node in converted_nodes.items():
            cases_ids = entry_addr_to_ids[entry_addr]
            if len(cases_ids) == 1:
                cases[next(iter(cases_ids))] = converted_node
            else:
                cases[tuple(sorted(cases_ids))] = converted_node

            self._new_sequences.append(converted_node)

        # reorganize cases to handle fallthroughs
        cases = self._reorganize_switch_cases(cases)

        return cases, node_default, to_remove

    #
    # Dealing with If-Then-Else structures
    #

    def _make_ites(self, seq):
        # search for a == ^a pairs

        while True:
            break_hard = False
            for i in range(len(seq.nodes)):
                node_0 = seq.nodes[i]
                if type(node_0) is not CodeNode:
                    continue
                rcond_0 = node_0.reaching_condition
                if rcond_0 is None:
                    continue
                if claripy.is_true(rcond_0) or claripy.is_false(rcond_0):
                    continue
                for j in range(i + 1, len(seq.nodes)):
                    node_1 = seq.nodes[j]
                    if type(node_1) is not CodeNode:
                        continue
                    if node_0 is node_1:
                        continue
                    rcond_1 = node_1.reaching_condition
                    if rcond_1 is None:
                        continue
                    cond_ = claripy.simplify(claripy.Not(rcond_0) == rcond_1)
                    if claripy.is_true(cond_):
                        # node_0 and node_1 should be structured using an if-then-else
                        self._make_ite(seq, node_0, node_1)
                        break_hard = True
                        break
                if break_hard:
                    break
            else:
                break

    def _structure_common_subexpression_conditions(self, seq):
        # use common subexpressions to structure nodes and create more if-then-else instances

        i = 0
        while i < len(seq.nodes) - 1:
            structured = False
            node_0 = seq.nodes[i]
            if not isinstance(node_0, CodeNode):
                i += 1
                continue
            rcond_0 = node_0.reaching_condition
            if rcond_0 is None:
                i += 1
                continue
            subexprs_0 = list(get_ast_subexprs(rcond_0))

            for common_subexpr in subexprs_0:
                if claripy.is_true(common_subexpr):
                    continue
                candidates = self._nodes_guarded_by_common_subexpr(seq, common_subexpr, i + 1)
                if candidates:
                    candidates.insert(0, (i, node_0, subexprs_0))
                    new_node = self._create_seq_node_guarded_by_common_subexpr(common_subexpr, candidates)
                    self._new_sequences.append(new_node)

                    # remove all old nodes and replace them with the new node
                    for idx, _, _ in candidates:
                        seq.nodes[idx] = None
                    seq.nodes[i] = CodeNode(new_node, common_subexpr)
                    seq.nodes = [n for n in seq.nodes if n is not None]
                    structured = True
                    break

            if not structured:
                i += 1

    @staticmethod
    def _nodes_guarded_by_common_subexpr(seq, common_subexpr, starting_idx):
        candidates = []

        if common_subexpr is claripy.true():
            return []
        for j, node_1 in enumerate(seq.nodes[starting_idx:]):
            rcond_1 = getattr(node_1, "reaching_condition", None)
            if rcond_1 is None:
                continue
            subexprs_1 = list(get_ast_subexprs(rcond_1))
            if any(subexpr_1 is common_subexpr for subexpr_1 in subexprs_1):
                # we found one!
                candidates.append((starting_idx + j, node_1, subexprs_1))

        return candidates

    @staticmethod
    def _create_seq_node_guarded_by_common_subexpr(common_subexpr, candidates):
        new_nodes = []

        for _, node, subexprs in candidates:
            # :)
            new_subexprs = [ex for ex in subexprs if ex is not common_subexpr]
            new_node = CodeNode(
                node.node,
                claripy.And(*new_subexprs),
            )
            new_nodes.append(new_node)

        return SequenceNode(None if not new_nodes else new_nodes[0].addr, nodes=new_nodes)

    def _replace_complex_reaching_conditions(self, seq: SequenceNode):
        for i in range(len(seq.nodes)):
            node = seq.nodes[i]

            if (
                isinstance(node, CodeNode)
                and node.reaching_condition is not None
                and node.reaching_condition.op == "Or"
                and node.node in self.cond_proc.guarding_conditions
            ):
                guarding_condition = self.cond_proc.guarding_conditions[node.node]
                # the op of guarding condition is always "Or"
                if (
                    len(guarding_condition.args) < len(node.reaching_condition.args)
                    and guarding_condition.depth < node.reaching_condition.depth
                ):
                    node.reaching_condition = guarding_condition

    def _make_condition_nodes(self, seq):
        # make all conditionally-reachable nodes ConditionNodes
        for i in range(len(seq.nodes)):
            node = seq.nodes[i]

            if isinstance(node, CodeNode):
                if isinstance(node.node, SequenceNode) and node.node not in self._new_sequences:
                    self._make_condition_nodes(node.node)

                if node.reaching_condition is not None and not claripy.is_true(node.reaching_condition):
                    if isinstance(node.node, ConditionalBreakNode):
                        # Put conditions together and simplify them
                        cond = claripy.And(node.reaching_condition, node.node.condition)
                        new_node = CodeNode(ConditionalBreakNode(node.node.addr, cond, node.node.target), None)
                    else:
                        new_node = ConditionNode(node.addr, None, node.reaching_condition, node, None)
                    seq.nodes[i] = new_node

    @staticmethod
    def _make_cascading_condition_nodes(seq: SequenceNode):
        """
        Convert nested condition nodes into a CascadingConditionNode.
        """
        CascadingConditionTransformer(seq)

    def _make_ite(self, seq, node_0, node_1):
        # ensure order
        if node_0.addr > node_1.addr:
            node_0, node_1 = node_1, node_0

        node_0_pos = seq.node_position(node_0)
        node_1_pos = seq.node_position(node_1)
        pos = max(node_0_pos, node_1_pos)

        node_0_, node_1_ = node_0.copy(), node_1.copy()
        # clear their reaching conditions
        node_0_.reaching_condition = None
        node_1_.reaching_condition = None

        node_0_kids = self._nodes_guarded_by_common_subexpr(seq, node_0.reaching_condition, node_0_pos + 1)
        node_0_kids.insert(0, (node_0_pos, node_0_, [node_0.reaching_condition]))
        node_1_kids = self._nodes_guarded_by_common_subexpr(seq, node_1.reaching_condition, node_1_pos + 1)
        node_1_kids.insert(0, (node_1_pos, node_1_, [node_1.reaching_condition]))

        new_node_0 = self._create_seq_node_guarded_by_common_subexpr(node_0.reaching_condition, node_0_kids)
        new_node_1 = self._create_seq_node_guarded_by_common_subexpr(node_1.reaching_condition, node_1_kids)

        self._new_sequences.append(new_node_0)
        self._new_sequences.append(new_node_1)

        seq_addr = seq.addr

        # erase all nodes in the candidates
        for idx, _, _ in node_0_kids + node_1_kids:
            seq.nodes[idx] = None

        seq.insert_node(pos, ConditionNode(seq_addr, None, node_0.reaching_condition, new_node_0, new_node_1))
        seq.nodes = [n for n in seq.nodes if n is not None]


# delayed import
from angr.analyses.decompiler.sequence_walker import SequenceWalker  # pylint:disable=wrong-import-position



# pylint:disable=missing-class-docstring,no-self-use
from __future__ import annotations
import logging

from angr.analyses import ForwardAnalysis, visitors
from angr.block import SootBlockNode
from angr.errors import AngrLoopAnalysisError
from . import register_analysis
from .analysis import Analysis


l = logging.getLogger(name=__name__)


class VariableTypes:
    Iterator = "Iterator"
    HasNext = "HasNext"
    Next = "Next"


class AnnotatedVariable:
    __slots__ = ["type", "variable"]

    def __init__(self, variable, type_):
        self.variable = variable
        self.type = type_

    def __repr__(self):
        return f"{self.variable}//{self.type}"

    def __eq__(self, other):
        return type(other) is AnnotatedVariable and other.type == self.type and other.variable == self.variable


class Condition:
    Equal = "=="
    NotEqual = "!="

    def __init__(self, op, val0, val1):
        self.op = op
        self.val0 = val0
        self.val1 = val1

    def __repr__(self):
        return f"{self.val0} {self.op} {self.val1}"

    @classmethod
    def from_opstr(cls, opstr):
        mapping = {
            "eq": cls.Equal,
            "==": cls.Equal,
            "ne": cls.NotEqual,
            "!=": cls.NotEqual,
        }

        return mapping.get(opstr)


class SootBlockProcessor:
    def __init__(self, state, block, loop, defuse):
        self.state = state
        self.block = block
        self.loop = loop
        self.defuse = defuse

    def process(self):
        if not isinstance(self.block, SootBlockNode):
            raise AngrLoopAnalysisError(f"Got an unexpected type of block {type(self.block)}.")

        if not self.block.stmts:
            return None

        for stmt in self.block.stmts:
            func_name = f"_handle_{stmt.__class__.__name__}"

            if hasattr(self, func_name):
                getattr(self, func_name)(stmt)

        return self.state

    def _stmt_inside_loop(self, stmt_idx):
        """
        Test whether a statement is inside the loop body or not.

        :param stmt_idx:
        :return:
        """

        # TODO: This is slow. Fix the performance issue

        return any(node.addr.stmt_idx <= stmt_idx < node.addr.stmt_idx + node.size for node in self.loop.body_nodes)

    def _expr(self, expr):
        func_name = f"_handle_{expr.__class__.__name__}"

        if hasattr(self, func_name):
            return getattr(self, func_name)(expr)

        return expr

    #
    # Statement handlers
    #

    def _handle_AssignStmt(self, stmt):
        left_op, right_op = stmt.left_op, stmt.right_op

        expr = self._expr(right_op)
        if expr is not None:
            try:
                from pysoot.sootir.soot_value import SootLocal
            except ImportError:
                l.error("Please install PySoot before analyzing Java byte code.")
                raise
            if isinstance(left_op, SootLocal):
                # Log a def for the local variable
                self.state.locals[left_op.name] = expr

    def _handle_IfStmt(self, stmt):
        target = stmt.target

        # is it jumping outside the loop?
        if not self._stmt_inside_loop(target):
            cond = Condition(
                Condition.from_opstr(stmt.condition.op),
                self._expr(stmt.condition.value1),
                self._expr(stmt.condition.value2),
            )

            self.state.add_loop_exit_stmt(stmt.label, condition=cond)

    #
    # Expression handlers
    #

    def _handle_SootLocal(self, expr):
        local_name = expr.name

        # First try state.locals
        try:
            return self.state.locals[local_name]
        except KeyError:
            pass

        # Then try defuse graph
        try:
            def_ = self.defuse.defs[local_name]
            expr = def_.evaluated
        except KeyError:
            pass

        return expr

    def _handle_SootIntConstant(self, expr):
        return expr.value

    def _handle_SootInterfaceInvokeExpr(self, expr):
        full_method = expr.class_name + "." + expr.method_name

        mapping = {
            "java.util.Set.iterator": VariableTypes.Iterator,
            "java.util.Iterator.hasNext": VariableTypes.HasNext,
            "java.util.Iterator.next": VariableTypes.Next,
        }

        base_var = self._expr(expr.base)

        if base_var is None:
            return None

        # Try to annotate the variable when applicable
        try:
            try:
                from pysoot.sootir.soot_value import SootValue
            except ImportError:
                l.error("Please install PySoot before analyzing Java byte code.")
                raise

            var_type = mapping[full_method]

            if isinstance(base_var, (SootValue, AnnotatedVariable)):
                return AnnotatedVariable(base_var, var_type)

        except KeyError:
            pass

        return None


class LoopAnalysisState:
    def __init__(self, block):
        self.block = block

        self.induction_variables = {}
        self.locals = {}

        self.loop_exit_stmts = set()

    def __repr__(self):
        return f"<LoopAnalysisState {self.block.addr}>"

    def copy(self):
        state = LoopAnalysisState(block=self.block)

        state.induction_variables = self.induction_variables.copy()
        state.locals = self.locals.copy()
        state.loop_exit_stmts = self.loop_exit_stmts.copy()

        return state

    def merge(self, state):
        s = self.copy()

        # TODO: Induction variables

        s.locals.update(state.locals)
        s.loop_exit_stmts |= state.loop_exit_stmts

        return s

    def add_loop_exit_stmt(self, stmt_idx, condition=None):
        self.loop_exit_stmts.add((condition, stmt_idx))


class LoopAnalysis(ForwardAnalysis, Analysis):
    """
    Analyze a loop and recover important information about the loop (e.g., invariants, induction variables) in a static
    manner.
    """

    def __init__(self, loop, defuse):
        visitor = visitors.LoopVisitor(loop)
        ForwardAnalysis.__init__(self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=visitor)

        self.loop = loop
        self.defuse = defuse
        self._traversed = set()
        self._last_state = None

        self.loop_exit_stmts = None
        self.locals = None
        self.bounded = None  # Whether this loop is bounded

        self._analyze()

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        pass

    def _pre_job_handling(self, job):
        pass

    def _initial_abstract_state(self, node):
        return LoopAnalysisState(node)

    def _merge_states(self, node, *states):
        merged = states[0]
        for other in states[1:]:
            if other is not None:
                merged = merged.merge(other)
        return merged

    def _run_on_node(self, node, state):
        if node in self._traversed:
            return False, state

        self._traversed.add(node)

        state = state.copy()

        processor = SootBlockProcessor(state, node, self.loop, self.defuse)
        state = processor.process()

        self._last_state = state

        return True, state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        self.loop_exit_stmts = self._last_state.loop_exit_stmts
        self.locals = self._last_state.locals

        self._last_state = None

        # Is it bounded?
        self.bounded = self._is_bounded()

    def _is_bounded(self):
        """
        Checks whether this loop is bounded. We basically does a bunch of pattern matching.

        :return: True if this loop is bounded, False is this loop is not bounded, None otherwise (undetermined).
        :rtype: bool or None
        """

        b = self._is_bounded_iterator_based()
        if b is not None:
            return b

        # Cannot determine
        return None

    def _is_bounded_iterator_based(self):
        """
        Iterator based check.

        With respect to a certain variable/value A,
        - there must be at least one exit condition being A//Iterator//HasNext == 0
        - there must be at least one local that ticks the iterator next: A//Iterator//Next
        """

        # Condition 0
        def check_0(cond):
            return (
                isinstance(cond, Condition)
                and cond.op == Condition.Equal
                and cond.val1 == 0
                and isinstance(cond.val0, AnnotatedVariable)
                and cond.val0.type == VariableTypes.HasNext
            )

        check_0_results = [(check_0(stmt[0]), stmt[0]) for stmt in self.loop_exit_stmts]
        check_0_conds = [cond for r, cond in check_0_results if r]  # remove all False ones

        if not check_0_conds:
            return None

        the_iterator = check_0_conds[0].val0.variable

        # Condition 1
        def check_1(local):
            return (
                isinstance(local, AnnotatedVariable)
                and local.type == VariableTypes.Next
                and local.variable == the_iterator
            )

        if not any(check_1(local) for local in self.locals.values()):
            return None

        return True


register_analysis(LoopAnalysis, "LoopAnalysis")



# pylint:disable=import-outside-toplevel
from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable, Iterable
import queue
import threading
import time
import logging
from collections import defaultdict
from enum import Enum

import networkx

import claripy

from angr.utils.graph import GraphUtils
from angr.simos import SimWindows
from angr.utils.mp import mp_context, Initializer
from angr.knowledge_plugins.cfg import CFGModel
from . import Analysis, register_analysis, VariableRecoveryFast, CallingConventionAnalysis, CFGFast

if TYPE_CHECKING:
    from angr.calling_conventions import SimCC
    from angr.sim_type import SimTypeFunction
    from angr.knowledge_plugins.variables.variable_manager import VariableManagerInternal
    from angr.knowledge_plugins.functions.function_manager import Function


_l = logging.getLogger(name=__name__)

_mp_context = mp_context()


class CallingConventionAnalysisMode(Enum):
    """
    The mode of calling convention analysis.

    FAST: Using FactCollector to collect facts, then use facts for calling convention analysis.
    VARIABLES: Using variables in VariableManager for calling convention analysis.
    """

    FAST = "fast"
    VARIABLES = "variables"


class CompleteCallingConventionsAnalysis(Analysis):
    """
    Implements full-binary calling convention analysis. During the initial analysis of a binary, you may set
    `recover_variables` to True so that it will perform variable recovery on each function before performing calling
    convention analysis.
    """

    def __init__(
        self,
        mode: CallingConventionAnalysisMode = CallingConventionAnalysisMode.FAST,
        recover_variables=False,
        low_priority=False,
        force=False,
        cfg: CFGFast | CFGModel | None = None,
        analyze_callsites: bool = False,
        skip_signature_matched_functions: bool = False,
        max_function_blocks: int | None = None,
        max_function_size: int | None = None,
        workers: int = 0,
        cc_callback: Callable | None = None,
        prioritize_func_addrs: list[int] | set[int] | None = None,
        skip_other_funcs: bool = False,
        auto_start: bool = True,
        func_graphs: dict[int, networkx.DiGraph] | None = None,
    ):
        """

        :param recover_variables:   Recover variables on each function before performing calling convention analysis.
        :param low_priority:        Run in the background - periodically release GIL.
        :param force:               Perform calling convention analysis on functions even if they have calling
                                    conventions or prototypes already specified (or previously recovered).
        :param cfg:                 The control flow graph model, which will be passed to CallingConventionAnalysis.
        :param analyze_callsites:   Consider artifacts at call sites when performing calling convention analysis.
        :param skip_signature_matched_functions:    Do not perform calling convention analysis on functions that match
                                                    against existing FLIRT signatures.
        :param max_function_blocks: Do not perform calling convention analysis on functions with more than the
                                    specified number of blocks. Setting it to None disables this check.
        :param max_function_size:   Do not perform calling convention analysis on functions whose sizes are more than
                                    `max_function_size`. Setting it to None disables this check.
        :param workers:             Number of multiprocessing workers.
        """

        self.mode = mode
        self._recover_variables = recover_variables
        self._low_priority = low_priority
        self._force = force
        self._cfg: CFGModel | None = cfg.model if isinstance(cfg, CFGFast) else cfg
        self._analyze_callsites = analyze_callsites
        self._skip_signature_matched_functions = skip_signature_matched_functions
        self._max_function_blocks = max_function_blocks
        self._max_function_size = max_function_size
        self._workers = workers
        self._cc_callback = cc_callback
        self._prioritize_func_addrs = prioritize_func_addrs
        self._skip_other_funcs = skip_other_funcs
        self._auto_start = auto_start
        self._total_funcs = None
        self._func_graphs = func_graphs if func_graphs else {}
        self.prototype_libnames: set[str] = set()

        # sanity check
        if self.mode not in {CallingConventionAnalysisMode.FAST, CallingConventionAnalysisMode.VARIABLES}:
            raise ValueError(f"Invalid calling convention analysis mode {self.mode}.")

        self._func_addrs = []  # a list that holds addresses of all functions to be analyzed
        self._results = []
        if workers > 0:
            self._remaining_funcs = _mp_context.Value("i", 0)
            self._results = _mp_context.Queue()
            self._results_lock = _mp_context.Lock()
            self._func_queue = _mp_context.Queue()
            self._func_queue_lock = _mp_context.Lock()
        else:
            self._remaining_funcs = None  # not needed
            self._results_lock = None  # not needed
            self._func_queue = None  # not needed
            self._func_queue_lock = threading.Lock()

        self._analyze()
        if self._auto_start:
            self.work()

    def _analyze(self):
        """
        Infer calling conventions for all functions in the current project.
        """

        # special case: if both _prioritize_func_addrs and _skip_other_funcs are set, we only need to sort part of
        # the call graph; even better, if there is only one function set, we don't need to sort the call graph at all!
        if self._prioritize_func_addrs and self._skip_other_funcs:
            if len(self._prioritize_func_addrs) == 1:
                self._func_addrs = list(self._prioritize_func_addrs)
                self._total_funcs = 1
                return
            directed_callgraph = networkx.DiGraph(self.kb.functions.callgraph)
            directed_callgraph = directed_callgraph.subgraph(self._prioritize_func_addrs)
        else:
            # get an ordering of functions based on the call graph
            # note that the call graph is a multi-digraph. we convert it to a digraph to speed up topological sort
            directed_callgraph = networkx.DiGraph(self.kb.functions.callgraph)
        assert isinstance(directed_callgraph, networkx.DiGraph)
        sorted_funcs = GraphUtils.quasi_topological_sort_nodes(directed_callgraph)

        total_funcs = 0
        for func_addr in reversed(sorted_funcs):
            func = self.kb.functions.get_by_addr(func_addr)
            if (func.calling_convention is None or func.prototype is None) or self._force:
                if func.is_alignment:
                    # skip all alignments
                    continue

                if self._skip_signature_matched_functions and func.from_signature:
                    # this function matches against a known library function. skip it.
                    continue

                if self._max_function_size is not None:
                    func_size = sum(block.size for block in func.blocks if block.size is not None)
                    if func_size > self._max_function_size:
                        _l.info(
                            "Skipping variable recovery for %r since its size (%d) is greater than the cutoff "
                            "size (%d).",
                            func,
                            func_size,
                            self._max_function_size,
                        )
                        continue

                if self._max_function_blocks is not None and len(func.block_addrs_set) > self._max_function_blocks:
                    _l.info(
                        "Skipping variable recovery for %r since its number of blocks (%d) is greater than the "
                        "cutoff number (%d).",
                        func,
                        len(func.block_addrs_set),
                        self._max_function_blocks,
                    )
                    continue

                # if it's a normal function, we attempt to perform variable recovery
                self._func_addrs.append(func_addr)
                total_funcs += 1

        self._total_funcs = total_funcs

        if self._prioritize_func_addrs:
            self.prioritize_functions(self._prioritize_func_addrs)
        self._prioritize_func_addrs = None  # no longer useful

    def _set_function_prototype(
        self,
        func: Function,
        prototype: SimTypeFunction | None,
        prototype_libname: str | None,
        prototype_guessed: bool | None,
    ) -> None:
        if func.prototype is None or func.is_prototype_guessed or self._force:
            if prototype_guessed is not None:
                func.is_prototype_guessed = prototype_guessed
            func.prototype = prototype
            func.prototype_libname = prototype_libname

    def work(self):
        total_funcs = self._total_funcs
        assert total_funcs is not None
        if self._workers == 0:
            self._update_progress(0)
            for idx, func_addr in enumerate(self._func_addrs):
                cc, proto, proto_libname, proto_guessed, _ = self._analyze_core(func_addr)

                func = self.kb.functions.get_by_addr(func_addr)
                if cc is not None or proto is not None:
                    func.calling_convention = cc
                    self._set_function_prototype(func, proto, proto_libname, proto_guessed)
                    if proto_libname is not None:
                        self.prototype_libnames.add(proto_libname)
                func.ran_cca = True

                if self._cc_callback is not None:
                    self._cc_callback(func_addr)

                percentage = (idx + 1) / total_funcs * 100.0
                self._update_progress(percentage, text=f"{idx + 1}/{total_funcs} - {func.demangled_name}")
                if self._low_priority:
                    self._release_gil(idx + 1, 10, 0.000001)
            self._finish_progress()

        else:
            assert self._remaining_funcs is not None and self._func_queue is not None
            self._remaining_funcs.value = len(self._func_addrs)

            # generate a call tree (obviously, it's acyclic)
            traversed_func_addrs = set()
            depends_on = {}
            dependents = defaultdict(set)
            func_addrs_set = set(self._func_addrs)
            for func_addr in reversed(self._func_addrs):
                traversed_func_addrs.add(func_addr)
                depends_on[func_addr] = set()
                for callee in self.kb.functions.callgraph.successors(func_addr):
                    if callee not in traversed_func_addrs and callee in func_addrs_set:
                        depends_on[func_addr].add(callee)
                        dependents[callee].add(func_addr)

            # enqueue all leaf functions
            for func_addr in [k for k in depends_on if not depends_on[k]]:  # pylint:disable=consider-using-dict-items
                self._func_queue.put((func_addr, None))
                del depends_on[func_addr]

            self._update_progress(0, text="Spawning workers...")
            cc_callback = self._cc_callback
            self._cc_callback = None

            if self.project.simos is not None and isinstance(self.project.simos, SimWindows):
                # delayed import
                from angr.procedures.definitions import load_win32api_definitions

                Initializer.get().register(load_win32api_definitions)

            # spawn workers to perform the analysis
            with self._func_queue_lock:
                procs = [
                    _mp_context.Process(target=self._worker_routine, args=(worker_id, Initializer.get()), daemon=True)
                    for worker_id in range(self._workers)
                ]
                for proc_idx, proc in enumerate(procs):
                    self._update_progress(0, text=f"Spawning worker {proc_idx}...")
                    proc.start()

            self._cc_callback = cc_callback

            # update progress
            self._update_progress(0)
            idx = 0
            assert self._results_lock is not None
            while idx < total_funcs:
                try:
                    with self._results_lock:
                        func_addr, cc, proto, proto_libname, proto_guessed, varman = self._results.get(
                            True, timeout=0.01
                        )
                except queue.Empty:
                    time.sleep(0.1)
                    continue

                func = self.kb.functions.get_by_addr(func_addr)
                if cc is not None or proto is not None:
                    func.calling_convention = cc
                    self._set_function_prototype(func, proto, proto_libname, proto_guessed)
                    if proto_libname is not None:
                        self.prototype_libnames.add(proto_libname)

                if varman is not None:
                    self.kb.variables.function_managers[func_addr] = varman
                    varman.set_manager(self.kb.variables)
                func.ran_cca = True

                if self._cc_callback is not None:
                    self._cc_callback(func_addr)

                idx += 1

                percentage = idx / total_funcs * 100.0
                self._update_progress(percentage, text=f"{idx}/{total_funcs} - {func.demangled_name}")
                if self._low_priority:
                    self._release_gil(idx, 10, 0.0000001)

                # enqueue functions whose callees have been analyzed
                if func_addr in dependents:
                    for dependent in dependents[func_addr]:
                        depends_on[dependent].discard(func_addr)
                        if not depends_on[dependent]:
                            callee_prototypes = self._get_callees_cc_prototypes(dependent)
                            with self._func_queue_lock:
                                self._func_queue.put((dependent, callee_prototypes))
                            del depends_on[dependent]

            for proc in procs:
                proc.join()

            self._finish_progress()

    def _worker_routine(self, worker_id: int, initializer: Initializer):
        initializer.initialize()
        idx = 0
        assert self._remaining_funcs is not None and self._func_queue is not None and self._results_lock is not None
        while self._remaining_funcs.value > 0:
            try:
                with self._func_queue_lock:
                    func_addr, callee_info = self._func_queue.get(True, timeout=0.01)
                    self._remaining_funcs.value -= 1
            except queue.Empty:
                time.sleep(0.3)
                continue

            if callee_info is not None:
                callee_info: dict[int, tuple[SimCC | None, SimTypeFunction | None, str | None, bool | None]]
                for callee, (
                    callee_cc,
                    callee_proto,
                    callee_proto_libname,
                    callee_proto_guessed,
                ) in callee_info.items():
                    callee_func = self.kb.functions.get_by_addr(callee)
                    callee_func.calling_convention = callee_cc
                    self._set_function_prototype(callee_func, callee_proto, callee_proto_libname, callee_proto_guessed)

            idx += 1
            if self._low_priority and idx % 3 == 0:
                time.sleep(0.1)

            try:
                cc, proto, proto_libname, proto_guessed, varman = self._analyze_core(func_addr)
            except Exception:  # pylint:disable=broad-except
                _l.error("Worker %d: Exception occurred during _analyze_core().", worker_id, exc_info=True)
                cc, proto, proto_libname, proto_guessed, varman = None, None, None, None, None
            with self._results_lock:
                self._results.put((func_addr, cc, proto, proto_libname, proto_guessed, varman))

    def _analyze_core(
        self, func_addr: int
    ) -> tuple[SimCC | None, SimTypeFunction | None, str | None, bool | None, VariableManagerInternal | None]:
        func = self.kb.functions.get_by_addr(func_addr)
        if func.ran_cca:
            return (
                func.calling_convention,
                func.prototype,
                func.prototype_libname,
                func.is_prototype_guessed,
                self.kb.variables.get_function_manager(func_addr),
            )

        if (
            self.mode == CallingConventionAnalysisMode.VARIABLES
            and self._recover_variables
            and self.function_needs_variable_recovery(func)
        ):
            # special case: we don't have a PCode-engine variable recovery analysis for PCode architectures!
            if ":" in self.project.arch.name and self._func_graphs and func.addr in self._func_graphs:
                # this is a pcode architecture
                return None, None, None, None, None

            _l.info("Performing variable recovery on %r...", func)
            try:
                _ = self.project.analyses[VariableRecoveryFast].prep(kb=self.kb)(
                    func, low_priority=self._low_priority, func_graph=self._func_graphs.get(func.addr, None)
                )
            except claripy.ClaripyError:
                _l.warning(
                    "An claripy exception occurred during variable recovery analysis on function %#x.",
                    func.addr,
                    exc_info=True,
                )
                return None, None, None, None, None

        # determine the calling convention of each function
        cc_analysis = self.project.analyses[CallingConventionAnalysis].prep(kb=self.kb)(
            func,
            cfg=self._cfg,
            analyze_callsites=self._analyze_callsites,
            collect_facts=self.mode == CallingConventionAnalysisMode.FAST,
        )

        if cc_analysis.cc is not None:
            _l.info("Determined calling convention and prototype for %r.", func)
            return (
                cc_analysis.cc,
                cc_analysis.prototype,
                cc_analysis.prototype_libname if cc_analysis.prototype_libname is not None else func.prototype_libname,
                not cc_analysis.proto_from_symbol,
                self.kb.variables.get_function_manager(func_addr),
            )
        _l.info("Cannot determine calling convention for %r.", func)
        return None, None, None, None, self.kb.variables.get_function_manager(func_addr)

    def prioritize_functions(self, func_addrs_to_prioritize: Iterable[int]):
        """
        Prioritize the analysis of specified functions.

        :param func_addrs_to_prioritize: A collection of function addresses to analyze first.
        """

        with self._func_queue_lock:
            func_addrs_to_prioritize = set(func_addrs_to_prioritize)
            to_prioritize = []
            remaining = []
            for addr in self._func_addrs:
                if addr in func_addrs_to_prioritize:
                    to_prioritize.append(addr)
                else:
                    if not self._skip_other_funcs:
                        remaining.append(addr)

            self._func_addrs = to_prioritize + remaining

    def _get_callees_cc_prototypes(
        self, caller_func_addr: int
    ) -> dict[int, tuple[SimCC | None, SimTypeFunction | None, str | None, bool | None]]:
        d = {}
        for callee in self.kb.functions.callgraph.successors(caller_func_addr):
            if callee != caller_func_addr and callee not in d:
                func = self.kb.functions.get_by_addr(callee)
                tpl = func.calling_convention, func.prototype, func.prototype_libname, func.is_prototype_guessed
                d[callee] = tpl
        return d

    #
    # Static methods
    #

    @staticmethod
    def function_needs_variable_recovery(func):
        """
        Check if running variable recovery on the function is the only way to determine the calling convention of the
        this function.

        We do not need to run variable recovery to determine the calling convention of a function if:
        - The function is a SimProcedure.
        - The function is a PLT stub.
        - The function is a library function and we already know its prototype.

        :param func:    The function object.
        :return:        True if we must run VariableRecovery before we can determine what the calling convention of this
                        function is. False otherwise.
        :rtype:         bool
        """

        # TODO: Check SimLibraries
        return not (func.is_simprocedure or func.is_plt)


register_analysis(CompleteCallingConventionsAnalysis, "CompleteCallingConventions")
